;; Generated package description from jabber.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "jabber" "0.12.5.0.20260801.12" "XMPP/Jabber client" '((emacs "29.1") (fsm "0.2.0") (keymap-popup "0.2")) :commit "2cbf8764759a4765393578dca821b2d5234672ad" :authors '(("Magnus Henoch" . "mange@freemail.hu")) :maintainer '("Thanos Apollo" . "public@thanosapollo.org") :keywords '("comm") :url "https://git.thanosapollo.org/emacs-jabber/about/")
